/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "gstxma.h"

#include "gstxrm.h"

#include <gio/gio.h>
#include <xma.h>

GST_DEBUG_CATEGORY_STATIC (gst_xma_debug);
#define GST_CAT_DEFAULT gst_xma_debug

enum
{
  PROP_0,
  PROP_DEVICE,
};

static void gst_xma_initable_iface_init (gpointer g_iface);

G_DEFINE_TYPE_WITH_CODE (GstXma, gst_xma, G_TYPE_OBJECT,
    G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, gst_xma_initable_iface_init));

static void
gst_xma_set_property (GObject * object, guint prop_id, const GValue * value,
    GParamSpec * pspec)
{
  GstXma *self = GST_XMA (object);

  switch (prop_id) {
    case PROP_DEVICE:
      self->device = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
_xma_log_cb (void * opaque, XmaLogLevelType level, const char * name,
    const char * msg)
{
  GstDebugLevel gst_level = GST_LEVEL_NONE;

  switch (level) {
    case XMA_EMERGENCY_LOG:
      gst_level = GST_LEVEL_ERROR;
      break;
    case XMA_ALERT_LOG:
      gst_level = GST_LEVEL_ERROR;
      break;
    case XMA_CRITICAL_LOG:
      gst_level = GST_LEVEL_ERROR;
      break;
    case XMA_ERROR_LOG:
      gst_level = GST_LEVEL_ERROR;
      break;
    case XMA_WARNING_LOG:
      gst_level = GST_LEVEL_WARNING;
      break;
    case XMA_NOTICE_LOG:
      gst_level = GST_LEVEL_INFO;
      break;
    case XMA_INFO_LOG:
      gst_level = GST_LEVEL_INFO;
      break;
    case XMA_DEBUG_LOG:
      gst_level = GST_LEVEL_DEBUG;
      break;
  }

  /* Treat XmaLogLevelType as GstDebugLevel. */
  GST_CAT_LEVEL_LOG (GST_CAT_DEFAULT, gst_level, NULL, "[%s] %s", name, msg);
}

G_LOCK_DEFINE_STATIC (xma_devices);
static GSList *xma_devices = NULL;
static XmaLogHandle log_handle = NULL;

XmaLogHandle
gst_xma_get_log_handle ()
{
  if (!log_handle) {
    if (xma_log_init (XMA_DEBUG_LOG, XMA_LOG_TYPE_CALLBACK, &log_handle,
        _xma_log_cb, NULL) != XMA_SUCCESS) {
      GST_ERROR ("xma_log_init() failed");
    }

    GST_DEBUG ("Initialized XMA logger");
  }

  return log_handle;
}

static gboolean
gst_xma_initable_init (GInitable * initable, GCancellable * cancellable,
    GError ** error)
{
  GstXma *self = GST_XMA (initable);

  gboolean result = FALSE;
  XmaInitParameter param;
  gchar *app_name = gst_version_string ();

  param.app_name = app_name;
  param.device = self->device;
  param.params = NULL;
  param.param_cnt = 0;

  if (xma_initialize (gst_xma_get_log_handle (), &param, &self->xma) == XMA_SUCCESS) {
    GST_DEBUG ("Initialized XMA library for device %d", self->device);
    result = TRUE;
  } else {
    GST_ERROR ("Failed to initialize XMA library for device %d", param.device);
    result = FALSE;
  }

  g_clear_pointer (&app_name, g_free);

  return result;
}

static void
gst_xma_initable_iface_init (gpointer g_iface)
{
  GInitableIface *iface = g_iface;

  iface->init = gst_xma_initable_init;
}

static void
gst_xma_init (GstXma * xma)
{
}

static void
gst_xma_finalize (GObject * object)
{
  GstXma *self = GST_XMA (object);

  xma_release (self->xma);
  GST_DEBUG ("Released XMA library for device %d", self->device);
}

static void
gst_xma_class_init (GstXmaClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->set_property = gst_xma_set_property;
  gobject_class->finalize = gst_xma_finalize;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_uint ("device", "XMA device",
          "Numeric ID of an XMA device", 0, G_MAXUINT32, 0,
          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));

  GST_DEBUG_CATEGORY_INIT (gst_xma_debug, "xma", 0, "Xilinx XMA interface");
}

static void
gst_xma_weak_ref_notify (gpointer unused, GObject * xma)
{
  G_LOCK (xma_devices);
  xma_devices = g_slist_remove (xma_devices, xma);

  if (!xma_devices) {
    xma_log_release (g_steal_pointer (&log_handle));
    GST_DEBUG ("Released XMA logger");
  }

  G_UNLOCK (xma_devices);
}

GstXma *
gst_xma_get_for_device (guint32 device_id)
{
  GstXma *xma = NULL;
  GSList *it;

  device_id = gst_xrm_choose_device (device_id);
  if (device_id == GST_XRM_ERROR) {
    goto error;
  }

  G_LOCK (xma_devices);

  for (it = xma_devices; it; it = g_slist_next (it)) {
    GstXma *x = GST_XMA (it->data);
    if (x->device == device_id) {
      xma = g_object_ref (x);
      break;
    }
  }

  if (!xma) {
    if (xma_devices) {
      GST_ERROR ("Using more than one XMA device per process is not allowed");
    } else {
      xma = g_initable_new (GST_TYPE_XMA, NULL, NULL, "device", device_id, NULL);
      if (xma) {
        g_object_weak_ref (G_OBJECT (xma), gst_xma_weak_ref_notify, NULL);
        xma_devices = g_slist_append (xma_devices, xma);
      }
    }
  }

  G_UNLOCK (xma_devices);

error:
  return xma;
}

XmaFormatType
gst_xma_get_format (GstVideoFormat fmt, gint32 *flags)
{
  *flags = 0;

  switch (fmt) {
    case GST_VIDEO_FORMAT_NV12:
      return XMA_NV12_FMT_TYPE;
    case GST_VIDEO_FORMAT_NV12_4L4:
      *flags = XMA_FRAME_PROPERTY_FLAG_TILE_4x4;
      return XMA_NV12_FMT_TYPE;
    case GST_VIDEO_FORMAT_NV21:
      return XMA_NV21_FMT_TYPE;
    case GST_VIDEO_FORMAT_P010_10LE:
      return XMA_P010LE_FMT_TYPE;
    case GST_VIDEO_FORMAT_NV12_10LE_4L4:
      *flags = XMA_FRAME_PROPERTY_FLAG_TILE_4x4;
      return XMA_P010LE_FMT_TYPE;
    case GST_VIDEO_FORMAT_P010_10BE:
      return XMA_P010BE_FMT_TYPE;
    case GST_VIDEO_FORMAT_I420:
      return XMA_YUV420P_FMT_TYPE;
    case GST_VIDEO_FORMAT_I420_10LE:
      return XMA_YUV420P10LE_FMT_TYPE;
    case GST_VIDEO_FORMAT_I420_10BE:
      return XMA_YUV420P10BE_FMT_TYPE;
    case GST_VIDEO_FORMAT_Y42B:
      return XMA_YUV422P_FMT_TYPE;
    case GST_VIDEO_FORMAT_I422_10LE:
      return XMA_YUV422P10LE_FMT_TYPE;
    case GST_VIDEO_FORMAT_I422_10BE:
      return XMA_YUV422P10BE_FMT_TYPE;
    case GST_VIDEO_FORMAT_Y444:
      return XMA_YUV444P_FMT_TYPE;
    case GST_VIDEO_FORMAT_ARGB:
      return XMA_ARGB_FMT_TYPE;
    case GST_VIDEO_FORMAT_ABGR:
      return XMA_ABGR_FMT_TYPE;
    case GST_VIDEO_FORMAT_RGBA:
      return XMA_RGBA_FMT_TYPE;
    case GST_VIDEO_FORMAT_BGRA:
      return XMA_BGRA_FMT_TYPE;
    case GST_VIDEO_FORMAT_RGB:
      return XMA_RGB24_FMT_TYPE;
    case GST_VIDEO_FORMAT_BGR:
      return XMA_BGR24_FMT_TYPE;
    /* XMA_VPE_FMT_TYPE */
    case GST_VIDEO_FORMAT_NV12_10LE32:
      return XMA_PACKED10_FMT_TYPE;
    case GST_VIDEO_FORMAT_BGRx:
      return XMA_BGR0_FMT_TYPE;
    case GST_VIDEO_FORMAT_RGB16:
      return XMA_RGB16_FMT_TYPE;
    case GST_VIDEO_FORMAT_RGBP:
      return XMA_RGB24_P_FMT_TYPE;
    case GST_VIDEO_FORMAT_BGRP:
      return XMA_BGR24_P_FMT_TYPE;
    /* XMA_RGB48_FMT_TYPE */
    /* XMA_BGR48_FMT_TYPE */
    /* XMA_RGB48_P_FMT_TYPE */
    /* XMA_BGR48_P_FMT_TYPE */
    default:
      return XMA_NONE_FMT_TYPE;
  }
}

GstVideoFormat
gst_xma_from_format (XmaFormatType fmt, gint32 flags)
{
  if ((flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) && fmt != XMA_NV12_FMT_TYPE) {
    return GST_VIDEO_FORMAT_UNKNOWN;
  }

  switch (fmt) {
    case XMA_NV12_FMT_TYPE:
      return (flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) ?
          GST_VIDEO_FORMAT_NV12_4L4 : GST_VIDEO_FORMAT_NV12;
    case XMA_NV21_FMT_TYPE:
      return GST_VIDEO_FORMAT_NV21;
    case XMA_P010LE_FMT_TYPE:
      return (flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) ?
          GST_VIDEO_FORMAT_NV12_10LE_4L4 : GST_VIDEO_FORMAT_P010_10LE;
    case XMA_P010BE_FMT_TYPE:
      return GST_VIDEO_FORMAT_P010_10BE;
    case XMA_YUV420P_FMT_TYPE:
      return GST_VIDEO_FORMAT_I420;
    case XMA_YUV420P10LE_FMT_TYPE:
      return GST_VIDEO_FORMAT_I420_10LE;
    case XMA_YUV420P10BE_FMT_TYPE:
      return GST_VIDEO_FORMAT_I420_10BE;
    case XMA_YUV422P_FMT_TYPE:
      return GST_VIDEO_FORMAT_Y42B;
    case XMA_YUV422P10LE_FMT_TYPE:
      return GST_VIDEO_FORMAT_I422_10LE;
    case XMA_YUV422P10BE_FMT_TYPE:
      return GST_VIDEO_FORMAT_I422_10BE;
    case XMA_YUV444P_FMT_TYPE:
      return GST_VIDEO_FORMAT_Y444;
    case XMA_ARGB_FMT_TYPE:
      return GST_VIDEO_FORMAT_ARGB;
    case XMA_ABGR_FMT_TYPE:
      return GST_VIDEO_FORMAT_ABGR;
    case XMA_RGBA_FMT_TYPE:
      return GST_VIDEO_FORMAT_RGBA;
    case XMA_BGRA_FMT_TYPE:
      return GST_VIDEO_FORMAT_BGRA;
    case XMA_RGB24_FMT_TYPE:
      return GST_VIDEO_FORMAT_RGB;
    case XMA_BGR24_FMT_TYPE:
      return GST_VIDEO_FORMAT_BGR;
    /* XMA_VPE_FMT_TYPE */
    case XMA_PACKED10_FMT_TYPE:
      return GST_VIDEO_FORMAT_NV12_10LE32;
    case XMA_BGR0_FMT_TYPE:
      return GST_VIDEO_FORMAT_BGRx;
    case XMA_RGB16_FMT_TYPE:
      return GST_VIDEO_FORMAT_RGB16;
    case XMA_RGB24_P_FMT_TYPE:
      return GST_VIDEO_FORMAT_RGBP;
    case XMA_BGR24_P_FMT_TYPE:
      return GST_VIDEO_FORMAT_BGRP;
    /* XMA_RGB48_FMT_TYPE */
    /* XMA_BGR48_FMT_TYPE */
    /* XMA_RGB48_P_FMT_TYPE */
    /* XMA_BGR48_P_FMT_TYPE */
    default:
      return GST_VIDEO_FORMAT_UNKNOWN;
  }
}
